<?php

/**
 * @file
 * Field module related hook implementations for GitLab Field feature.
 */

/**
 * Implements hook_field_info().
 */
function gitlab_field_field_info() {
  return array(
    'gitlab_field' => array(
      'label' => t('GitLab file'),
      'description' => t('Attach GitLab files to entities.'),
      'settings' => array(),
      'instance_settings' => array(),
      'default_widget' => 'gitlab_field',
      'default_formatter' => 'gitlab_field_default',
    ),
  );
}

/**
 * Implements hook_field_widget_info().
 */
function gitlab_field_field_widget_info() {
  return array(
    'gitlab_field' => array(
      'label' => t('GitLab file'),
      'field types' => array('gitlab_field'),
      'settings' => array(
        'content_display' => array(
          'allowed_file_extensions' => '',
        ),
        'gitlab_field_defaults' => array(
          'default_account' => 0,
          'default_project' => 0,
          'default_branch' => 'master',
        ),
      ),
      'behaviors' => array(
        'multiple values' => FIELD_BEHAVIOR_CUSTOM,
        'default value' => FIELD_BEHAVIOR_NONE,
      ),
    ),
  );
}

/**
 * Adds default GitLab field settings widget to any form, Field API or anything.
 *
 * @param $form
 *   The form to be altered. It'll have a 'gitlab_field_defaults' fieldset.
 * @param $defaults
 *   The defaults to be loaded to that form as an array with the following keys:
 *   - default_account: integer. The remote ID of the default GitLab account.
 *   - default_project: integer. The remote ID of the default GitLab project.
 *   - default_branch: string. The remote ID of the default GitLab branch.
 */
function _gitlab_field_form_alter_defaults(&$form, $defaults) {
  // Load available GitLab accounts.
  $form['#gitlab_accounts'] = gitlab_accounts_load_accounts();

  $form['gitlab_field_defaults'] = array(
    '#type' => 'fieldset',
    '#title' => t('Default GitLab field settings'),
    '#collapsible' => FALSE,
    '#collapsed' => FALSE,
  );

  // Widget settings forms don't have a form_id set at this time. Other forms
  // will need this #tree set to TRUE, though, to be able to handle both cases
  // the same way during AJAX calls.
  if (!empty($form['#form_id'])) {
    $form['gitlab_field_defaults']['#tree'] = TRUE;
  }

  $form['gitlab_field_defaults']['default_account'] = array(
    '#type' => 'select',
    '#title' => t('Default account'),
    '#options' => array(),
    '#default_value' => $defaults['default_account'] ? $defaults['default_account'] : 0,
    '#empty_option' => t('- Select -'),
    '#disabled' => count($form['#gitlab_accounts']) == 0,
    '#ajax' => array(
      'callback' => '_gitlab_field_field_gitlab_field_widget_settings_form_default_account_ajax_callback',
      'effect' => 'fade',
    ),
  );

  $form['gitlab_field_defaults']['default_project'] = array(
    '#type' => 'select',
    '#title' => t('Default project'),
    '#options' => array(),
    '#default_value' => $defaults['default_project'] ? $defaults['default_project'] : 0,
    '#empty_option' => t('- Select -'),
    '#disabled' => count($form['#gitlab_accounts']) == 0,
    '#ajax' => array(
      'callback' => '_gitlab_field_field_gitlab_field_widget_settings_form_default_project_ajax_callback',
      'effect' => 'fade',
    ),
    // Preserve us from AJAX validation errors.
    '#validated' => TRUE,
  );

  $form['gitlab_field_defaults']['default_branch'] = array(
    '#type' => 'select',
    '#title' => t('Default branch'),
    '#options' => array(),
    '#default_value' => $defaults['default_branch'] ? $defaults['default_branch'] : '',
    '#empty_option' => t('- Select -'),
    '#disabled' => count($form['#gitlab_accounts']) == 0,
    // Preserve us from AJAX validation errors.
    '#validated' => TRUE,
  );

  // Add our defaults to the form so after_build functions will be able to reuse
  // them.
  $form['#gitlab_field_defaults'] = $defaults;

  $form['#after_build'][] = 'gitlab_field_field_gitlab_field_widget_settings_form_after_build';
}

/**
 * Implements hook_widget_settings_form().
 */
function gitlab_field_field_widget_settings_form($field, $instance) {
  $widget = $instance['widget'];
  $settings = $widget['settings'];

  $form = array();
  if ($widget['type'] == 'gitlab_field') {
    $form['content_display'] = array(
      '#type' => 'fieldset',
      '#title' => t('File content display'),
      '#collapsible' => FALSE,
      '#collapsed' => FALSE,
    );

    $form['content_display']['allowed_file_extensions'] = array(
      '#type' => 'textfield',
      '#title' => t('Allowed file extensions'),
      '#description' => t('The space separated file extension list the content display will be available for.'),
      '#default_value' => $settings['content_display']['allowed_file_extensions'],
      '#element_validate' => array('gitlab_field_widget_allowed_file_extensions_validate'),
    );

    // Add the default selection widget.
    _gitlab_field_form_alter_defaults($form, $settings['gitlab_field_defaults']);
    // Storage is already handled automatically by the Field API.
  }

  return $form;
}

/**
 * Custom form after build function.
 *
 * @see gitlab_field_field_widget_settings_form()
 */
function gitlab_field_field_gitlab_field_widget_settings_form_after_build($form, &$form_state) {
  $form['#gitlab_account'] = NULL;

  if (isset($form['#gitlab_field_defaults']['default_account']) && array_key_exists($form['#gitlab_field_defaults']['default_account'], $form['#gitlab_accounts'])) {
    $form['#gitlab_account'] = gitlab_accounts_load_account($form['#gitlab_field_defaults']['default_account']);
  }
  elseif (isset($form['gitlab_field_defaults']['default_account']['#value']) && array_key_exists($form['gitlab_field_defaults']['default_account']['#value'], $form['#gitlab_accounts'])) {
    $form['#gitlab_account'] = gitlab_accounts_load_account($form['gitlab_field_defaults']['default_account']['#value']);
  }

  $account_options = array();
  $project_options = array();
  $branch_options = array();

  array_walk($form['#gitlab_accounts'], function ($account) use (&$account_options) {
    $account_options[$account->id] = check_plain($account->name . ' (' . $account->username . ')');
  });

  // If the Gitlab account set, then try to load its projects.
  if ($form['#gitlab_account']) {
    $project_options += _gitlab_field_get_project_field_options($form['#gitlab_account']);

    if (!empty($project_options) && array_key_exists($form['#gitlab_field_defaults']['default_project'], $project_options)) {
      $branch_options += _gitlab_field_get_branch_field_options($form['#gitlab_account'], $form['#gitlab_field_defaults']['default_project']);
    }
  }

  $form['gitlab_field_defaults']['default_account']['#options'] += $account_options;
  $form['gitlab_field_defaults']['default_project']['#options'] += $project_options;
  $form['gitlab_field_defaults']['default_branch']['#options'] += $branch_options;

  return $form;
}

/**
 * AJAX callback for Default account field.
 *
 * @see gitlab_field_field_widget_settings_form()
 */
function _gitlab_field_field_gitlab_field_widget_settings_form_default_account_ajax_callback($form, $form_state) {
  $commands = array();
  if ($form['#form_id'] == 'field_ui_field_edit_form') {
    // When we're on a field's widget settings form (ie. field edit form),
    // retrieve the form part for the defaults from the field instance settings
    // and prepare the selector as needed.
    $ws_form = &$form['instance']['widget']['settings'];
    $selector_prefix = '.form-item-instance-widget-settings-gitlab-field-defaults-';
  }
  else {
    // When we're on a different form, assume our own default selector widget
    // has been added accordingly.
    $ws_form = &$form;
    $selector_prefix = '.form-item-gitlab-field-defaults-';
  }

  $ws_form['gitlab_field_defaults']['default_project']['#options'] = array(0 => t('- Select -'));
  $ws_form['gitlab_field_defaults']['default_branch']['#options'] = array('' => t('- Select -'));

  if ($ws_form['#gitlab_account']) {
    $project_options = _gitlab_field_get_project_field_options($ws_form['#gitlab_account']);
    $ws_form['gitlab_field_defaults']['default_project']['#options'] += $project_options;
  }

  $commands[] = ajax_command_replace($selector_prefix . 'default-project', render($ws_form['gitlab_field_defaults']['default_project']));
  $commands[] = ajax_command_replace($selector_prefix . 'default-branch', render($ws_form['gitlab_field_defaults']['default_branch']));

  return array(
    '#type' => 'ajax',
    '#commands' => $commands,
  );
}

/**
 * AJAX callback for Default project field.
 *
 * @see gitlab_field_field_widget_settings_form()
 */
function _gitlab_field_field_gitlab_field_widget_settings_form_default_project_ajax_callback($form, $form_state) {
  $commands = array();
  if ($form['#form_id'] == 'field_ui_field_edit_form') {
    // When we're on a field's widget settings form (ie. field edit form),
    // retrieve the form part for the defaults from the field instance settings
    // and prepare the selector as needed.
    $ws_form = &$form['instance']['widget']['settings'];
    $ws_values = $form_state['values']['instance']['widget']['settings'];
    $selector_prefix = '.form-item-instance-widget-settings-gitlab-field-defaults-';
  }
  else {
    // When we're on a different form, assume our own default selector widget
    // has been added accordingly.
    $ws_form = &$form;
    $ws_values = $form_state['values'];
    $selector_prefix = '.form-item-gitlab-field-defaults-';
  }

  $ws_form['gitlab_field_defaults']['default_branch']['#options'] = array('' => t('- Select -'));
  if ($ws_form['#gitlab_account'] && $ws_values['gitlab_field_defaults']['default_project']) {
    $ws_form['gitlab_field_defaults']['default_branch']['#options'] += _gitlab_field_get_branch_field_options($ws_form['#gitlab_account'], $ws_values['gitlab_field_defaults']['default_project']);
  }

  $commands[] = ajax_command_replace($selector_prefix . 'default-branch', render($ws_form['gitlab_field_defaults']['default_branch']));

  return array(
    '#type' => 'ajax',
    '#commands' => $commands,
  );
}

/**
 * Element validate callback for GitLab field widget's settings form.
 *
 * @see gitlab_field_field_widget_settings_form()
 */
function gitlab_field_widget_allowed_file_extensions_validate($element, &$form_state, $form) {
  if (!preg_match('/^([a-z]+( [a-z]+)?)*$/', $element['#value'])) {
    form_error($element, t('File extensions must be in lowercase format and separated by space.'));
  }
}

/**
 * Implements hook_field_formatter_info().
 */
function gitlab_field_field_formatter_info() {
  $info = array(
    'gitlab_field_default' => array(
      'label' => t('Link'),
      'field types' => array('gitlab_field'),
      'description' => t('Link to GitLab files.'),
      'settings' => array(
        'content_display' => array(
          'text_format' => 'plain_text',
        ),
      ),
    ),
  );
  return $info;
}

/**
 * Implements hook_field_formatter_settings_summary().
 */
function gitlab_field_field_formatter_settings_summary($field, $instance, $view_mode) {
  $display = $instance['display'][$view_mode];
  $settings = $display['settings'];

  $summary = '';
  if ($display['type'] == 'gitlab_field_default') {
    $summary = t('<strong>Content display</strong><p>Text format: @text_format</p>', array(
      '@text_format' => empty($settings['text_format']) ? t('None') : $settings['content_display']['text_format'],
    ));
  }

  return $summary;
}

/**
 * Implements hook_field_formatter_settings_form().
 */
function gitlab_field_field_formatter_settings_form($field, $instance, $view_mode, $form, &$form_state) {
  $display = $instance['display'][$view_mode];
  $settings = $display['settings'];

  $form = array();
  if ($display['type'] == 'gitlab_field_default') {
    $form['content_display'] = array(
      '#type' => 'fieldset',
      '#title' => t('File content display'),
      '#collapsible' => FALSE,
      '#collapsed' => FALSE,
    );

    $text_format_options = array();
    $text_formats = filter_formats();
    foreach ($text_formats as $key => $text_format) {
      $text_format_options[$key] = $text_format->name;
    }

    $form['content_display']['text_format'] = array(
      '#type' => 'select',
      '#title' => t('Text format'),
      '#description' => t('The text format used to display file contents.'),
      '#options' => $text_format_options,
      '#default_value' => $settings['content_display']['text_format'],
      '#required' => TRUE,
    );
  }

  return $form;
}

/**
 * Implements hook_field_formatter_view().
 */
function gitlab_field_field_formatter_view($entity_type, $entity, $field, $instance, $langcode, $items, $display) {
  $element = array();
  $settings = $display['settings'];

  switch ($display['type']) {
    case 'gitlab_field_default':
      if (!empty($items)) {
        // Prepare the container for the files.
        $element[0] = array(
          '#type' => 'container',
          // Attributes property is needed for the bootstrap theme. Without it
          // the bootstrap theme will cause a fatal PHP error during the
          // rendering of container elements (on Example node view pages)
          // because it assumes that #attributes property is always exist.
          '#attributes' => array(),
          'file_contents' => array(),
          'table' => array(
            '#theme' => 'table',
            '#header' => array(
              t('File name'),
              t('Raw URL'),
              t('Updated'),
            ),
            '#rows' => array(),
            '#caption' => NULL,
            '#colgroups' => array(),
            '#sticky' => FALSE,
            '#empty' => NULL,
          ),
        );

        $item_file_ids = array();
        foreach ($items as $item) {
          $item_file_ids[] = $item['fid'];
        }
        // Get the referenced file objects.
        $file_objects = entity_load('gitlab_field_file', $item_file_ids);
        // Sort the items according to their weights.
        uasort($items, '_gitlab_field_widget_sort_field_items');

        // Walk through to items and display their properties.
        foreach ($items as $delta => $item) {
          if ($item['display']) {
            // Item related file object.
            $file_object = $file_objects[$item['fid']];
            $file_content = base64_decode($file_object->content);
            if (empty($file_content)) {
              $file_content = t('Empty file');
            }
            $data = array(
              'file_content' => $file_content,
              'text_format' => $settings['content_display']['text_format'],
            );
            // Enable modules to alter the file content and the text format.
            drupal_alter('gitlab_field_file_content_display_format', $data, $file_object);

            // Create a "cleared" file name for anchors.
            $cleared_file_name = trim($file_object->name);
            $cleared_file_name = preg_replace('/[^A-Za-z0-9 _\.]/', '', $cleared_file_name);
            $cleared_file_name = preg_replace('/ /', '-', $cleared_file_name);
            $cleared_file_name = strtolower($cleared_file_name);
            $cleared_file_name = $file_object->id . '-' . $cleared_file_name;

            $element[0]['file_contents'][] = array(
              '#type' => 'item',
              '#title' => l($file_object->name, NULL, array(
                'attributes' => array('name' => $cleared_file_name),
                'fragment' => $cleared_file_name,
                'external' => TRUE,
              )),
              'file_content' => array(
                '#markup' => check_markup($data['file_content'], $data['text_format']),
              ),
            );
          }
        }

        // Walk through to items and display their properties.
        foreach ($items as $delta => $item) {
          // Item related file object.
          $file_object = $file_objects[$item['fid']];
          $element[0]['table']['#rows'][] = array(
            l(check_plain($file_object->name), $file_object->getBlobURL(), array(
              'attributes' => array('target' => '_blank'),
            )),
            l(check_plain($file_object->name), $file_object->getRawURL(), array(
              'attributes' => array('target' => '_blank'),
            )),
            format_date($file_object->updated, 'custom', 'Y-m-d g:i a'),
          );
        }
      }
      break;

  }

  return $element;
}

/**
 * An uasort() callback function to sort field items by their weights.
 */
function _gitlab_field_widget_sort_field_items($item_a, $item_b) {
  if ($item_a['weight'] == $item_b['weight']) {
    return 0;
  }
  return ($item_a['weight'] < $item_b['weight']) ? -1 : 1;
}

/**
 * Implements hook_field_is_empty().
 */
function gitlab_field_field_is_empty($item, $field) {
  switch ($field['type']) {
    case 'gitlab_field':
      return empty($item['fid']);
      break;

  }
  return FALSE;
}

/**
 * Implements hook_field_presave().
 */
function gitlab_field_field_presave($entity_type, $entity, $field, $instance, $langcode, &$items) {
  switch ($field['type']) {
    case 'gitlab_field':
      foreach ($items as $item) {
        if (!empty($item['display'])) {
          // Try to load the referenced GitLab file object.
          $file_objects = entity_load('gitlab_field_file', [$item['fid']]);
          if (!empty($file_objects[$item['fid']])) {
            // Try to fetch the file's content from GitLab.
            $file_objects[$item['fid']]->fetchContent();
          }
        }
      }
      break;
  }
}

/**
 * Implements hook_field_widget_form().
 */
function gitlab_field_field_widget_form(&$form, &$form_state, $field, $instance, $langcode, $items, $delta, $element) {
  $settings = $instance['widget']['settings'];

  // Make sure we are working with the right widget.
  if ($instance['widget']['type'] == 'gitlab_field') {
    global $user;
    // Load the per-user defaults.
    $user_defaults = _gitlab_field_user_defaults_load($user->uid);

    $field_name = $field['field_name'];
    $field_lang = $langcode;

    // Element defaults.
    $element += array(
      '#type' => 'fieldset',
      '#collapsible' => FALSE,
      '#collapsed' => FALSE,
      '#element_validate' => array(
        '_gitlab_field_widget_validate',
      ),
      '#attributes' => array(
        'class' => array(
          'gitlab-file-field-widget',
          $field_name,
        ),
      ),
      '#attached' => array(
        'js' => array(
          drupal_get_path('module', 'gitlab_field') . '/gitlab_field.widget.js',
        ),
        'css' => array(
          drupal_get_path('module', 'gitlab_field') . '/gitlab_field.widget.css',
        ),
      ),
    );

    // Load available GitLab accounts.
    $accounts = gitlab_accounts_load_accounts();
    // Get context from $form_state.
    $context = _gitlab_field_widget_get_context($form_state, $field, $field_lang);
    // Extract context variables.
    list($account, $project, $branch, $folder_path, $files_in_folder, $files_to_link, $files_to_unlink, $files_weights) = $context;

    // Store field items under a field specific $form_state key.
    if (!isset($form_state[$field_name . '_gitlab_field_items'])) {
      $form_state[$field_name . '_gitlab_field_items'] = array();
      $form_state[$field_name . '_gitlab_field_items_file_objects'] = array();
      if (!empty($items)) {
        $item_ids = array();
        $field_items = array();
        foreach ($items as $item) {
          $item_ids[] = $item['fid'];
          $field_items[$item['fid']] = $item;
        }
        // Load file objects referenced in items.
        $file_objects = entity_load('gitlab_field_file', $item_ids);
        // Sort items by their weights.
        uasort($field_items, '_gitlab_field_widget_sort_field_items');
        // Walk through on the sorted field items and file objects.
        foreach ($field_items as $fid => $item) {
          $form_state[$field_name . '_gitlab_field_items'][$fid] = $field_items[$fid];
          $form_state[$field_name . '_gitlab_field_items_file_objects'][$fid] = $file_objects[$fid];
        }
      }
    }

    // Get a reference variable to field items stored under $form_state.
    $field_items = &$form_state[$field_name . '_gitlab_field_items'];
    $field_items_file_objects = &$form_state[$field_name . '_gitlab_field_items_file_objects'];

    // Check for new weights.
    if (!empty($files_weights)) {
      // Sort the fid => weight values by weight.
      asort($files_weights);
      // Sort field items and file objects according to the new weights.
      $field_items_sorted = array();
      $field_items_file_objects_sorted = array();
      foreach ($files_weights as $fid => $weight) {
        $field_items_sorted[$fid] = $field_items[$fid];
        $field_items_file_objects_sorted[$fid] = $field_items_file_objects[$fid];
      }
      $field_items = $field_items_sorted;
      $field_items_file_objects = $field_items_file_objects_sorted;
    }

    // Link new files.
    if (!empty($files_to_link)) {
      // Set a boolean to show a warning message about temporarily changes.
      _gitlab_field_widget_set_changes_warning($field_name, $form_state);

      $info = array(
        'account' => $account,
        'project' => $project,
        'branch' => $branch,
        'folder_path' => $folder_path,
        'field_items' => $field_items,
        'field_items_file_objects' => $field_items_file_objects,
      );
      _gitlab_field_link_new_files($files_to_link, $info, $field_items, $field_items_file_objects);
    }

    // Unlink files.
    foreach ($field_items as $fid => $file) {
      if (in_array($fid, $files_to_unlink)) {
        // Set a boolean to show a warning message about temporarily changes.
        _gitlab_field_widget_set_changes_warning($field_name, $form_state);
        // TODO: This should be corrected over time.
        unset($field_items[$fid]);
        unset($field_items_file_objects[$fid]);
      }
    }

    // Warning message container for temporarily changes.
    $element['changes_warning'] = array(
      '#type' => 'container',
      '#attributes' => array(
        'id' => $field_name . '_gitlab_field_widget_changes_warning',
        'style' => 'display: none;',
      ),
      'message' => array(
        '#theme' => 'gitlab_field_widget_messages',
        '#messages' => array(t('Changes will not be saved until the form is submitted.')),
        '#type' => 'warning',
      ),
    );

    $linked_files_params = array(
      'field' => $field,
      'language' => $field_lang,
      'field_items' => $field_items,
      'field_items_file_objects' => $field_items_file_objects,
      'widget_settings' => $instance['widget']['settings'],
    );
    $element['linked_files'] = _gitlab_field_widget_get_linked_files($linked_files_params);

    $element['file_browser_header'] = array(
      '#markup' => '<div style="margin:10px 0 20px 0;"><strong>' . t('File browser') . '</strong><hr/></div>',
    );

    // Construct unique HTML id for AJAX replace events.
    $link_new_files_html_id = $field_name . '_gitlab_field_widget_link_new_files';

    // Link new files container.
    $element['link_new_files'] = array(
      '#type' => 'container',
      '#attributes' => array(
        'id' => $link_new_files_html_id,
      ),
    );

    if (!empty($form_state['triggering_element']['#link_new_files'])) {
      // File sources container.
      $element['link_new_files']['file_sources'] = array(
        '#type' => 'container',
        '#attributes' => array(
          'class' => array(
            'file-sources',
            'container-inline',
          ),
        ),
      );

      $account_options = array();
      foreach ($accounts as $_account) {
        $account_options[$_account->id] = check_plain($_account->name . ' (' . $_account->username . ')');
      }

      // Initialize default account value.
      $default_account_value = '';

      // At least one account required.
      if (!empty($account_options)) {
        // If there is only one account select it by default.
        if (count($account_options) == 1) {
          reset($account_options);
          $account = $accounts[key($account_options)];
          _gitlab_field_widget_set_value($field_name, $form_state, 'account', $account);
          $default_account_value = $account->id;

          // Display the selected account.
          $element['link_new_files']['file_sources']['account'] = array(
            '#type' => 'select',
            '#title' => 'Account',
            '#options' => array(
              $account->id => check_plain($account->name . ' (' . $account->username . ')')
            ),
            '#default_value' => $account->id,
            '#disabled' => TRUE,
          );
        }
        else {
          if (empty($account)) {
            // If the field has Default account settings.
            if ($settings['gitlab_field_defaults']['default_account']) {
              $account = gitlab_accounts_load_account($settings['gitlab_field_defaults']['default_account']);
            }
            // Else if, if the user has Default account settings.
            elseif ($user_defaults['default_account']) {
              $account = gitlab_accounts_load_account($user_defaults['default_account']);
            }

            if (!empty($account)) {
              _gitlab_field_widget_set_value($field_name, $form_state, 'account', $account);
              $default_account_value = $account->id;
            }
          }

          // Account selection.
          $element['link_new_files']['file_sources']['account'] = array(
            '#type' => 'select',
            '#title' => 'Account',
            '#options' => $account_options,
            '#empty_option' => t('- Select -'),
            '#default_value' => $default_account_value,
            '#atttributes' => array('data-reset-value' => $default_account_value),
            '#link_new_files' => TRUE,
            // A custom property indicating what part of the form should be
            // validated when the element is the triggering element.
            '#file_sources_account_validate' => TRUE,
            '#widget_refresh' => array(
              'field' => array(
                'name' => $field_name,
                'language' => $field_lang,
              ),
              'replacements' => array(
                array(
                  'name_path' => 'link_new_files',
                  'html_id' => $link_new_files_html_id,
                ),
              ),
            ),
            '#ajax' => array(
              'callback' => '_gitlab_field_widget_refresh',
              'effect' => 'fade',
            ),
          );
        }

        // Reset the account selection value on client side if needed.
        if (!empty($form_state['triggering_element']['#reset_file_sources'])) {
          $element['link_new_files']['file_sources']['account']['#attributes'] = array(
            'data-reset-value' => $default_account_value,
          );
        }

        $project_options = array();
        if (!empty($account)) {
          $projects = gitlab_projects_index($account->getPrivateToken(), array(
            'wsdata' => array(
              'endpoint' => $account->endpoint_url,
            ),
          ));
          if (!empty($projects['data'])) {
            foreach ($projects['data'] as $_project) {
              $project_options[$_project['id']] = $_project['name_with_namespace'];
            }
            // No project selected yet, try to load a default one.
            if (empty($project)) {
              $_project_id = 0;
              // If the field has Default project settings.
              if ($settings['gitlab_field_defaults']['default_project'] && array_key_exists($settings['gitlab_field_defaults']['default_project'], $project_options)) {
                $_project_id = $settings['gitlab_field_defaults']['default_project'];
              }
              // Else if the user has Default project settings.
              elseif ($user_defaults['default_project'] && array_key_exists($user_defaults['default_project'], $project_options)) {
                $_project_id = $user_defaults['default_project'];
              }

              $project_data = gitlab_projects_read_project($_project_id, $account->getPrivateToken(), array(
                'wsdata' => array(
                  'endpoint' => $account->endpoint_url,
                ),
              ));
              if (!empty($project_data['data'])) {
                $project = $project_data['data'];
                _gitlab_field_widget_set_value($field_name, $form_state, 'project', $project_data['data']);
              }
            }
          }
        }

        $default_project_value = empty($project) ? '' : $project['id'];

        // Project selection.
        $element['link_new_files']['file_sources']['project'] = array(
          '#type' => 'select',
          '#title' => 'Project',
          '#options' => $project_options,
          '#empty_option' => t('- Select -'),
          '#default_value' => $default_project_value,
          '#attributes' => array(
            'data-reset-value' => $default_project_value,
          ),
          '#disabled' => empty($project_options),
          '#link_new_files' => TRUE,
          // A custom property indicating what part of the form should be
          // validated when the element is the triggering element.
          '#file_sources_project_validate' => TRUE,
          '#widget_refresh' => array(
            'field' => array(
              'name' => $field_name,
              'language' => $field_lang,
            ),
            'replacements' => array(
              array(
                'name_path' => 'link_new_files',
                'html_id' => $link_new_files_html_id,
              ),
            ),
          ),
          '#ajax' => array(
            'callback' => '_gitlab_field_widget_refresh',
            'effect' => 'fade',
          ),
        );

        // Reset the project selection value on client side if needed.
        if (!empty($form_state['triggering_element']['#reset_file_sources']) || empty($project)) {
          $element['link_new_files']['file_sources']['project']['#attributes'] = array(
            'data-reset-value' => $default_project_value,
          );
        }

        $branch_options = array();
        if (!empty($project)) {
          $branches = gitlab_projects_branches_index($project['id'], $account->getPrivateToken(), array(
            'wsdata' => array(
              'endpoint' => $account->endpoint_url,
            ),
          ));
          if (!empty($branches['data'])) {
            foreach ($branches['data'] as $_branch) {
              $branch_options[$_branch['name']] = $_branch['name'];
            }
          }
        }

        // Master branch should be the default branch.
        $default_branch_value = empty($branch) ? 'master' : $branch['name'];

        if (empty($branch)) {
          // If the field has Default branch settings.
          if ($settings['gitlab_field_defaults']['default_branch'] && array_key_exists($settings['gitlab_field_defaults']['default_branch'], $branch_options)) {
            $default_branch_value = $settings['gitlab_field_defaults']['default_branch'];
          }
          // Else if user has Default branch settings.
          elseif ($user_defaults['default_branch'] && array_key_exists($user_defaults['default_branch'], $branch_options)) {
            $default_branch_value = $user_defaults['default_branch'];
          }

          if (array_key_exists($default_branch_value, $branch_options)) {
            $branch_data = gitlab_projects_get_branch($project['id'], $default_branch_value, $account->getPrivateToken(), array(
              'wsdata' => array(
                'endpoint' => $account->endpoint_url,
              ),
            ));
            if (!empty($branch_data['data'])) {
              $branch = $branch_data['data'];
              _gitlab_field_widget_set_value($field_name, $form_state, 'branch', $branch);

              // Get the file list.
              $files_data = gitlab_repository_tree_index($project['id'], urlencode($folder_path), $branch['name'], $account->getPrivateToken(), array(
                'wsdata' => array(
                  'endpoint' => $account->endpoint_url,
                ),
              ));
              if (!empty($files_data['data'])) {
                $files_in_folder = $files_data['data'];
                _gitlab_field_widget_set_value($field_name, $form_state, 'files_in_folder', $files_in_folder);
              }
            }
          }
        }

        // Branch selection.
        $element['link_new_files']['file_sources']['branch'] = array(
          '#type' => 'select',
          '#title' => 'Branch',
          '#options' => $branch_options,
          '#empty_option' => t('- Select -'),
          '#default_value' => $default_branch_value,
          '#disabled' => empty($branch_options),
          '#attributes' => array(
            'data-reset-value' => $default_branch_value,
          ),
          '#link_new_files' => TRUE,
          // A custom property indicating what part of the form should be
          // validated when the element is the triggering element.
          '#file_sources_branch_validate' => TRUE,
          '#widget_refresh' => array(
            'field' => array(
              'name' => $field_name,
              'language' => $field_lang,
            ),
            'replacements' => array(
              array(
                'name_path' => 'link_new_files',
                'html_id' => $link_new_files_html_id,
              ),
            ),
          ),
          '#ajax' => array(
            'callback' => '_gitlab_field_widget_refresh',
            'effect' => 'fade',
          ),
        );

        // Reset the branch selection value on client side if needed.
        if (!empty($form_state['triggering_element']['#reset_file_sources']) || empty($branch)) {
          $element['link_new_files']['file_sources']['branch']['#attributes']['data-reset-value'] = $default_branch_value;
        }

        // File sources reset button.
        $element['link_new_files']['file_sources']['reset'] = array(
          '#type' => 'button',
          '#value' => t('Reset'),
          '#name' => $field_name . '_gitlab_field_widget_reset_file_sources_button',
          '#link_new_files' => TRUE,
          // Only validate the field the widget belongs to.
          '#limit_validation_errors' => array(array($field_name)),
          '#reset_file_sources' => TRUE,
          '#widget_refresh' => array(
            'field' => array(
              'name' => $field_name,
              'language' => $field_lang,
            ),
            'replacements' => array(
              array(
                'name_path' => 'link_new_files',
                'html_id' => $link_new_files_html_id,
              ),
            ),
          ),
          '#ajax' => array(
            'callback' => '_gitlab_field_widget_refresh',
            'effect' => 'fade',
          ),
        );

        // File sources reset button.
        $element['link_new_files']['file_sources']['refresh'] = array(
          '#type' => 'button',
          '#value' => t('Refresh'),
          '#name' => $field_name . '_gitlab_field_widget_refresh_file_sources_button',
          '#link_new_files' => TRUE,
          // Only validate the field the widget belongs to.
          '#limit_validation_errors' => array(array($field_name)),
          '#keep_folder_path' => TRUE,
          '#widget_refresh' => array(
            'field' => array(
              'name' => $field_name,
              'language' => $field_lang,
            ),
            'replacements' => array(
              array(
                'name_path' => 'link_new_files',
                'html_id' => $link_new_files_html_id,
              ),
            ),
          ),
          '#ajax' => array(
            'callback' => '_gitlab_field_widget_refresh',
            'effect' => 'fade',
          ),
        );

        if (!empty($account) && !empty($project) && !empty($branch)) {
          $file_browser_prams = array(
            'account' => $account,
            'project' => $project,
            'branch' => $branch,
            'field' => $field,
            'language' => $field_lang,
            'folder_path' => $folder_path,
            'files_in_folder' => $files_in_folder,
            'linked_files' => $field_items_file_objects,
          );

          // Get the file browser's render array.
          $element['link_new_files']['file_browser'] = _gitlab_field_widget_get_file_browser($file_browser_prams, $form_state);
        }
        else {
          $element['link_new_files']['file_browser'] = array(
            '#type' => 'container',
            '#attributes' => array(
              'id' => $field_name . '_gitlab_field_widget_file_browser',
              'class' => array('file-browser'),
            ),
            'message' => array(
              '#markup' => t('Select file source to list available files.'),
            ),
          );
        }
      }
      else {
        $element['link_new_files']['no_accounts'] = array(
          '#markup' => t('To link files you have to add at least one account on the GitLab accounts !configuration_page.', array(
            '!configuration_page' => l(t('configuration page'), 'admin/config/services/gitlab-file-field/accounts', array(
              'attributes' => array(
                'target' => '_blank',
              ),
            )),
          )),
        );
      }
    }
    else {
      $element['link_new_files']['link_new_files'] = array(
        '#type' => 'button',
        '#value' => t('Link new files'),
        '#link_new_files' => TRUE,
        '#widget_refresh' => array(
          'field' => array(
            'name' => $field_name,
            'language' => $field_lang,
          ),
          'replacements' => array(
            array(
              'name_path' => 'link_new_files',
              'html_id' => $link_new_files_html_id,
            ),
          ),
        ),
        '#ajax' => array(
          'callback' => '_gitlab_field_widget_refresh',
          'effect' => 'fade',
        ),
      );
    }
  }

  return $element;
}

/**
 * Element validate callback for file field widget.
 *
 * @see gitlab_field_field_widget_form()
 */
function _gitlab_field_widget_validate($element, &$form_state, $form) {
  $field_name = $element['#field_name'];
  $field_language = $element['#language'];

  // Reset file sources values if the related button pressed.
  if (!empty($form_state['triggering_element']['#reset_file_sources'])) {
    _gitlab_field_widget_set_value($field_name, $form_state, 'account', NULL);
    _gitlab_field_widget_set_value($field_name, $form_state, 'project', NULL);
    _gitlab_field_widget_set_value($field_name, $form_state, 'branch', NULL);
  }

  // Validate the selected account.
  if (!empty($form_state['triggering_element']['#file_sources_account_validate'])) {
    // Delete previously saved values from $form_state and validate them again.
    _gitlab_field_widget_set_value($field_name, $form_state, 'account', NULL);
    _gitlab_field_widget_set_value($field_name, $form_state, 'project', NULL);
    _gitlab_field_widget_set_value($field_name, $form_state, 'branch', NULL);
    // Check if $form_state value exists. It should.
    if (!empty($form_state['values'][$field_name][$field_language]['link_new_files']['file_sources']['account'])) {
      $account_id = $form_state['values'][$field_name][$field_language]['link_new_files']['file_sources']['account'];
      $loc_account = gitlab_accounts_load_account($account_id);
      // Try to connect to the selected GitLab account.
      $rem_account = gitlab_users_get_current_user($loc_account->getPrivateToken(), array(
        'wsdata' => array(
          'endpoint' => $loc_account->endpoint_url,
        ),
      ));
      if (empty($rem_account['data'])) {
        // TODO: this should be fixed.
        form_error($element['account']);
        _gitlab_field_widget_set_message($field_name, 'error', 'file_sources', t(
          'Cannot connect to the selected GitLab account. Please try again.'
        ));
      }
      else {
        _gitlab_field_widget_set_value($field_name, $form_state, 'account', $loc_account);
      }
    }
  }

  // Get the account from $form_state. It will be NULL without validation.
  $account = _gitlab_field_widget_get_value($field_name, $form_state, 'account');

  // Validate the selected project.
  if (!empty($form_state['triggering_element']['#file_sources_project_validate'])) {
    // Delete previously saved value from $form_state and validate it again.
    _gitlab_field_widget_set_value($field_name, $form_state, 'project', NULL);
    _gitlab_field_widget_set_value($field_name, $form_state, 'branch', NULL);
    // Check if account and $form_state value exists. They should.
    if (!empty($account) && !empty($form_state['values'][$field_name][$field_language]['link_new_files']['file_sources']['project'])) {
      $project_id = $form_state['values'][$field_name][$field_language]['link_new_files']['file_sources']['project'];
      // Try to connect to the selected project.
      $project_data = gitlab_projects_read_project($project_id, $account->getPrivateToken(), array(
        'wsdata' => array(
          'endpoint' => $account->endpoint_url,
        ),
      ));
      if (empty($project_data['data'])) {
        // TODO: this should be fixed.
        form_error($element['project']);
        _gitlab_field_widget_set_message($field_name, 'error', 'file_sources', t(
          'Cannot connect to the selected project. Please try again.'
        ));
      }
      else {
        _gitlab_field_widget_set_value($field_name, $form_state, 'project', $project_data['data']);
      }
    }
  }

  // Get the project from $form_state. It will be NULL without validation.
  $project = _gitlab_field_widget_get_value($field_name, $form_state, 'project');

  // Validate the selected branch.
  if (!empty($form_state['triggering_element']['#file_sources_branch_validate'])) {
    // Delete previously saved value from $form_state and validate it again.
    _gitlab_field_widget_set_value($field_name, $form_state, 'branch', NULL);
    // Check if project and $form_state value exists. They should.
    if (!empty($project) && !empty($form_state['values'][$field_name][$field_language]['link_new_files']['file_sources']['branch'])) {
      $branch_id = $form_state['values'][$field_name][$field_language]['link_new_files']['file_sources']['branch'];
      // Try to connect to the selected branch.
      $branch_data = gitlab_projects_get_branch($project['id'], $branch_id, $account->getPrivateToken(), array(
        'wsdata' => array(
          'endpoint' => $account->endpoint_url,
        ),
      ));
      if (empty($branch_data['data'])) {
        // TODO: this should be fixed.
        form_error($element['branch']);
        _gitlab_field_widget_set_message($field_name, 'error', 'file_sources', t(
          'Cannot connect to the selected branch. Please try again.'
        ));
      }
      else {
        _gitlab_field_widget_set_value($field_name, $form_state, 'branch', $branch_data['data']);
      }
    }
  }

  // Get the branch from $form_state. It will be NULL without validation.
  $branch = _gitlab_field_widget_get_value($field_name, $form_state, 'branch');

  $folder_path = NULL;
  // Field and widget specific "data-*" attribute for folder buttons.
  $folder_path_data_id = 'data-' . $field_name . '-folder-path';
  if (!empty($branch) && !empty($form_state['triggering_element']['#attributes'][$folder_path_data_id])) {
    $folder_path = $form_state['triggering_element']['#attributes'][$folder_path_data_id];
  }

  // Keep the saved folder path or refresh it with the new one.
  if (!empty($form_state['triggering_element']['#keep_folder_path'])) {
    $folder_path = _gitlab_field_widget_get_value($field_name, $form_state, 'folder_path');
  }
  else {
    _gitlab_field_widget_set_value($field_name, $form_state, 'folder_path', $folder_path);
  }

  // Validate field items' properties.
  if (!empty($form_state['values'][$field_name][$field_language])) {
    $weights_are_invalid = FALSE;
    $no_files_have_been_selected = TRUE;

    foreach (element_children($form_state['values'][$field_name][$field_language]) as $item_index) {
      $item = $form_state['values'][$field_name][$field_language][$item_index];
      if (empty($item['fid'])) {
        continue;
      }
      if (!empty($item['unlink'])) {
        $no_files_have_been_selected = FALSE;
      }
      if (!(ctype_digit($item['weight']) && (((int) $item['weight'] >= -9999) && ((int) $item['weight'] <= 9999)))) {
        form_error($element[$item_index]['weight']);
        $weights_are_invalid = TRUE;
      }
    }

    // Check if there are selected files when the unlink button pressed.
    $unlink_files_button_name = $field_name . '_gitlab_field_widget_unlink_files_button';
    $triggering_element_name = empty($form_state['triggering_element']['#name']) ? NULL : $form_state['triggering_element']['#name'];
    if (($triggering_element_name == $unlink_files_button_name) && $no_files_have_been_selected) {
      foreach (element_children($form_state['values'][$field_name][$field_language]) as $item_index) {
        $item = $form_state['values'][$field_name][$field_language][$item_index];
        if (empty($item['fid'])) {
          continue;
        }
        form_error($element[$item_index]['unlink']);
      }
      _gitlab_field_widget_set_message($field_name, 'error', 'linked_files', t(
        'No files have been selected.'
      ));
    }

    // Check if items' weights are valid.
    if ($weights_are_invalid) {
      _gitlab_field_widget_set_message($field_name, 'error', 'linked_files', t(
        'File weights must be integer values in the range [-9999, 9999].'
      ));
    }
  }

  // The file list in the current folder is always empty by default.
  _gitlab_field_widget_set_value($field_name, $form_state, 'files_in_folder', NULL);

  // Optionally refresh the file list in the selected folder.
  if (!empty($account) && !empty($project) && !empty($branch)) {
    $project_id = $project['id'];
    $branch_name = $branch['name'];
    $files_data = gitlab_repository_tree_index($project_id, urlencode($folder_path), $branch_name, $account->getPrivateToken(), array(
      'wsdata' => array(
        'endpoint' => $account->endpoint_url,
      ),
    ));
    if (!empty($files_data['data'])) {
      _gitlab_field_widget_set_value($field_name, $form_state, 'files_in_folder', $files_data['data']);
    }
  }

  $files_in_folder = _gitlab_field_widget_get_value($field_name, $form_state, 'files_in_folder');
  if ($files_in_folder == NULL) {
    $files_in_folder = array();
  }

  // Get the name of the files marked for link in the current folder.
  $link_files_button_name = $field_name . '_gitlab_field_widget_link_files_button';
  if (!empty($branch) && !empty($form_state['triggering_element']) && ($form_state['triggering_element']['#name'] == $link_files_button_name)) {
    $no_files_have_been_selected = TRUE;
    $rows = $form_state['values'][$field_name][$field_language]['link_new_files']['file_browser']['table']['rows'];
    foreach ($rows as $index => $value) {
      if (!empty($value['data']['file_name'])) {
        $no_files_have_been_selected = FALSE;

        $file_not_exists = TRUE;
        foreach ($files_in_folder as $file) {
          if (($file['type'] == 'blob') && ($file['name'] == $value['data']['file_name'])) {
            $file_not_exists = FALSE;
            break;
          }
        }
        if ($file_not_exists) {
          form_error($element['link_new_files']['file_browser']['table']['rows'][$index]['data']['file_name']);
          _gitlab_field_widget_set_message($field_name, 'error', 'file_browser', t(
            'The selected file @file_name cannot be linked because it is not found on the GitLab server.',
            array('@file_name' => $value['data']['file_name'])
          ));
        }
      }
    }
    if ($no_files_have_been_selected) {
      form_error($element['link_new_files']['file_browser']['table']);
      _gitlab_field_widget_set_message($field_name, 'error', 'file_browser', t(
        'No files have been selected.'
      ));
    }
  }
}

/**
 * Helper function to extract context variables from $form_state.
 *
 * @param array $form_state
 *   An associative array containing the current state of the form.
 * @param array $field
 *   An array of the field the widget belongs to.
 * @param string $language
 *   The language associated with field $items.
 *
 * @returns array
 *   An array of context variables.
 */
function _gitlab_field_widget_get_context(array &$form_state, array $field, $language) {
  $field_name = $field['field_name'];
  $field_lang = $language;

  $account = _gitlab_field_widget_get_value($field_name, $form_state, 'account');
  $project = _gitlab_field_widget_get_value($field_name, $form_state, 'project');
  $branch = _gitlab_field_widget_get_value($field_name, $form_state, 'branch');
  $folder_path = _gitlab_field_widget_get_value($field_name, $form_state, 'folder_path');
  $files_in_folder = _gitlab_field_widget_get_value($field_name, $form_state, 'files_in_folder');
  if ($files_in_folder == NULL) {
    $files_in_folder = array();
  }

  $files_weights = array();
  $files_to_unlink = array();
  if (!empty($form_state['values'][$field_name][$field_lang])) {
    foreach (element_children($form_state['values'][$field_name][$field_lang]) as $item_index) {
      $item = $form_state['values'][$field_name][$field_lang][$item_index];
      if (empty($item['fid'])) {
        continue;
      }
      if (!empty($item['unlink'])) {
        $files_to_unlink[] = $item['fid'];
      }
      $files_weights[$item['fid']] = $item['weight'];
    }
    // Get file ids which need to be unlinked.
    $unlink_files_button_name = $field_name . '_gitlab_field_widget_unlink_files_button';
    if (empty($form_state['triggering_element']) && ($form_state['triggering_element']['#name'] == $unlink_files_button_name)) {
      // Only collect unlinked files ids if the unlink related button pressed.
      $files_to_unlink = array();
    }
  }

  $files_to_link = array();
  // Get the name of the files marked for link in the current folder.
  $link_files_button_name = $field_name . '_gitlab_field_widget_link_files_button';
  if (!empty($branch) && !empty($form_state['triggering_element']) && ($form_state['triggering_element']['#name'] == $link_files_button_name)) {
    $rows = $form_state['values'][$field_name][$field_lang]['link_new_files']['file_browser']['table']['rows'];
    foreach ($rows as $value) {
      if (!empty($value['data']['file_name'])) {
        $files_to_link[] = $value['data']['file_name'];
      }
    }
  }

  return array(
    $account,
    $project,
    $branch,
    $folder_path,
    $files_in_folder,
    $files_to_link,
    $files_to_unlink,
    $files_weights,
  );
}

/**
 * Helper function to link new GitLab files.
 *
 * @param $file_names
 *   An array of file names in the currently selected folder.
 * @param $info
 *   Additional information for the linked files like account, project, etc.
 * @param $field_items
 *   An array of the current and sorted field items.
 * @param $field_items_file_objects
 *   An array of current and sorted file objects referenced by field items.
 *
 * @returns array
 *   Returns with the keyed array of the linked (saved) GitLab file entities.
 */
function _gitlab_field_link_new_files($file_names, $info, &$field_items, &$field_items_file_objects) {
  $account = $info['account'];
  $project = $info['project'];
  $branch = $info['branch'];

  $conditions = array(
    'endpoint_url' => $account->endpoint_url,
    'account_id' => $account->remote_id,
    'project_id' => $project['id'],
    'branch_id' => $branch['name'],
  );

  if (!empty($info['folder_path'])) {
    $conditions['folder_path'] = $info['folder_path'];
  }

  $existing_files = entity_load('gitlab_field_file', FALSE, $conditions);

  $existing_file_names = array();
  foreach ($existing_files as $file) {
    $existing_file_names[] = $file->name;
  }

  foreach ($file_names as $file_name) {
    // Only save the file entity if it doesn't exist.
    if (!in_array($file_name, $existing_file_names)) {
      // Create a new GitLab file entity.
      $new_file = entity_create('gitlab_field_file', array(
        'name' => $file_name,
        'folder_path' => $info['folder_path'],
        'endpoint_url' => $account->endpoint_url,
        'account_id' => $account->remote_id,
        'account_name' => $account->name,
        'project_id' => $project['id'],
        'project_name' => $project['name'],
        'project_url' => $project['web_url'],
        'branch_id' => $branch['name'],
        'content' => NULL,
        'created' => REQUEST_TIME,
        'updated' => REQUEST_TIME,
        'status' => 0,
      ));
      // Save the new entity.
      $new_file->save();
      // Add a new item to field items.
      $field_items[$new_file->id] = array(
        'fid' => $new_file->id,
        'display' => 0,
        'weight' => count($field_items),
      );
      // Add a new item to file objects.
      $field_items_file_objects[$new_file->id] = $new_file;
    }
    else {
      foreach ($existing_files as $existing_file) {
        if ($file_name == $existing_file->name) {
          // Add a new item to field items.
          $field_items[$existing_file->id] = array(
            'fid' => $existing_file->id,
            'display' => 0,
            'weight' => count($field_items),
          );
          // Add a new item to file objects.
          $field_items_file_objects[$existing_file->id] = $existing_file;
          break;
        }
      }
    }
  }
}

/**
 * Helper function to construct the linked files table.
 *
 * @param array $variables
 *   An associative array containing:
 *   - field: An array of the field the file browser widget belongs to.
 *   - language: The language associated with widget items.
 *   - field_items: The sorted field items belong to the field.
 *   - field_items_file_objects: The sorted and referenced file objects.
 *   - widget_settings: The widget settings array.
 *
 * @returns array
 *   The linked files table's render array.
 *
 * @see gitlab_field_field_widget_form()
 */
function _gitlab_field_widget_get_linked_files(array $variables) {
  $field_name = $variables['field']['field_name'];
  $field_lang = $variables['language'];
  $field_items = $variables['field_items'];
  $field_items_file_objects = $variables['field_items_file_objects'];
  $widget_settings = $variables['widget_settings'];
  // Create a unique HTML id for the file browser container.
  $container_html_id = $field_name . '_gitlab_field_widget_linked_files';

  if (!empty($field_items)) {
    $table_html_id = $field_name . '_gitlab_field_widget_linked_files_table';

    $table = array(
      '#theme' => 'gitlab_field_widget_linked_files_table',
      '#header' => array(
        t('File name'),
        t('Weight'),
        t('Folder'),
        t('Branch'),
        t('Project'),
        t('Account'),
        array(
          'data' => t('Display'),
          'width' => '20',
        ),
        array(
          'data' => t('Unlink'),
          'width' => '20',
        ),
      ),
      '#attributes' => array(
        'id' => $table_html_id,
        'class' => array(
          'gitlab_field_widget_linked_files_table',
          'linked-files',
        ),
        'data-changes-warning-id' => $field_name . '_gitlab_field_widget_changes_warning',
      ),
      '#caption' => NULL,
      '#colgroups' => array(),
      '#sticky' => FALSE,
      '#empty' => NULL,
      '#attached' => array(
        'drupal_add_tabledrag' => array(
          array($table_html_id, 'order', 'sibling', 'order-weight')
        ),
      ),
    );

    // Get content display related allowed file extensions from widget settings.
    $allowed_file_extensions = explode(' ', $widget_settings['content_display']['allowed_file_extensions']);

    // Rows added as child elements to the table.
    $row_index = 0;
    foreach ($field_items as $item) {
      // The file object by the item.
      $file_object = $field_items_file_objects[$item['fid']];
      $file_info = pathinfo($file_object->name);
      $file_extension = $file_info['extension'];

      // Construct the display content checkbox.
      if (in_array($file_extension, $allowed_file_extensions)) {
        $display_content = array(
          '#type' => 'checkbox',
          '#default_value' => $item['display'],
          '#attributes' => array(
            'class' => array('display'),
          ),
        );
      }
      else {
        $display_content = array(
          '#type' => 'checkbox',
          '#disabled' => TRUE,
        );
      }

      $table[$row_index] = array(
        '#parents' => array($field_name, $field_lang, $row_index),
        'fid' => array(
          '#type' => 'value',
          '#value' => $file_object->id,
        ),
        'file_name' => array(
          '#markup' => l(check_plain($file_object->name), $file_object->getBlobURL(), array(
            'attributes' => array('target' => '_blank'),
          )),
        ),
        'weight' => array(
          '#type' => 'textfield',
          '#title' => t('Weight'),
          '#title_display' => 'invisible',
          '#default_value' => $row_index,
          '#size' => 3,
          '#attributes' => array(
            'class' => array('order-weight'),
          ),
        ),
        'folder_path' => array(
          '#markup' => empty($file_object->folder_path) ? '/' : '/' . check_plain($file_object->folder_path),
        ),
        'branch_name' => array(
          '#markup' => check_plain($file_object->branch_id),
        ),
        'project_name' => array(
          '#markup' => check_plain($file_object->project_name),
        ),
        'account_name' => array(
          '#markup' => check_plain($file_object->account_name),
        ),
        'display' => $display_content,
        'unlink' => array(
          '#type' => 'checkbox',
        ),
      );

      $row_index++;
    }

    $unlink_files_button = array(
      '#type' => 'button',
      '#value' => t('Unlink selected files'),
      '#name' => $field_name . '_gitlab_field_widget_unlink_files_button',
      '#link_new_files' => TRUE,
      // Only validate the field the widget belongs to.
      '#limit_validation_errors' => array(array($field_name)),
      '#keep_folder_path' => TRUE,
      '#widget_refresh' => array(
        'field' => array(
          'name' => $field_name,
          'language' => $field_lang,
        ),
        'replacements' => array(
          array(
            'name_path' => 'linked_files',
            'html_id' => $container_html_id,
          ),
          array(
            'name_path' => 'link_new_files][file_browser',
            'html_id' => $field_name . '_gitlab_field_widget_file_browser',
          ),
        ),
      ),
      '#ajax' => array(
        'callback' => '_gitlab_field_widget_refresh',
        'effect' => 'fade',
      ),
    );

    $file_content_display_ext_list = '';
    $number_of_extensions = count($allowed_file_extensions);
    // List and highlight allowed file extensions for content display.
    if (!empty($allowed_file_extensions)) {
      foreach ($allowed_file_extensions as $index => $extension) {
        $file_content_display_ext_list .= '<strong>' . $extension . '</strong>';
        if ($index < ($number_of_extensions - 1)) {
          $file_content_display_ext_list .= ', ';
        }
      }
    }
    else {
      $file_content_display_ext_list = '<strong>none</strong>';
    }

    return array(
      '#type' => 'container',
      '#attributes' => array(
        'id' => $container_html_id,
      ),
      'messages' => array(
        '#type' => 'container',
        '#attributes' => array(
          'id' => $field_name . '_gitlab_field_widget_linked_files_messages',
        ),
      ),
      'file_content_display' => array(
        '#prefix' => '<div class="description">',
        '#markup' => t('The file content display functionality is available for the following file extensions: !extensions', array(
          '!extensions' => $file_content_display_ext_list,
        )),
        '#suffix' => '</div>',
      ),
      'table' => $table,
      'unlink_files' => $unlink_files_button,
    );
  }

  return array(
    '#type' => 'container',
    '#attributes' => array(
      'id' => $container_html_id,
    ),
    'no_files_linked' => array(
      '#markup' => t('No files have been linked yet.'),
    ),
  );
}

/**
 * Helper function to construct the file browser table.
 *
 * @param array $variables
 *   An associative array containing:
 *   - account: The selected GitLab account's object.
 *   - project: The selected GitLab project's property array.
 *   - branch: The selected GitLab branch's property array.
 *   - field: An array of the field the file browser widget belongs to.
 *   - language: The language associated with widget items.
 *   - folder_path: (optional) A folder path inside the project repository.
 *   - files_in_folder: The available file list in the selected folder path.
 *   - linked_files: The linked file objects belong to the field.
 * @param $form_state
 *   An associative array containing the current state of the form.
 *
 * @returns array
 *   The file browser's render array.
 *
 * @see gitlab_field_field_widget_form()
 */
function _gitlab_field_widget_get_file_browser(array $variables, &$form_state) {
  // Create a unique HTML id for the file browser container.
  $file_browser_html_id = $variables['field']['field_name'] . '_gitlab_field_widget_file_browser';

  // Boolean indicating that there are available linkable files or not.
  $linkable_files_available = FALSE;

  // The current folder path from the root of the project repository tree.
  $folder_path = '';
  // Field and widget specific "data-*" attribute name for folder buttons.
  $folder_path_data_name = 'data-' . $variables['field']['field_name'] . '-folder-path';

  $table_rows = array();
  $table_empty_message = t('No files to display.');

  $breadcrumb_button_elements = array();

  // Display the file list in a table.
  if (!empty($variables['files_in_folder'])) {
    // Init the file browser's breadcrumb with a button to the root folder.
    $button_params = array(
      'value' => check_plain($variables['project']['name_with_namespace'] . ' [' . $variables['branch']['name'] . ']'),
      'folder_path_data_name' => $folder_path_data_name,
      'folder_path' => NULL,
      'file_browser_html_id' => $file_browser_html_id,
      'field' => $variables['field'],
      'language' => $variables['language'],
    );
    // Construct a folder button according to the parameters.
    $breadcrumb_button_elements[] = _gitlab_field_widget_file_browser_folder_button($button_params);

    if ($variables['folder_path'] != '') {
      $folders = explode('/', $variables['folder_path']);
      while (!empty($folders)) {
        $folder = array_shift($folders);
        $folder_path .= empty($folder_path) ? $folder : ('/' . $folder);

        $button_params = array(
          'value' => $folder,
          'folder_path_data_name' => $folder_path_data_name,
          'folder_path' => $folder_path,
          'file_browser_html_id' => $file_browser_html_id,
          'field' => $variables['field'],
          'language' => $variables['language'],
        );
        // Construct a folder button according to the parameters.
        $breadcrumb_button_elements[] = _gitlab_field_widget_file_browser_folder_button($button_params);
      }
    }

    // Add a parent directory link to rows if the current path is not empty.
    if ($variables['folder_path'] != '') {
      $parent_folder_path = explode('/', $variables['folder_path']);
      array_pop($parent_folder_path);
      $parent_folder_path = implode('/', $parent_folder_path);

      $button_params = array(
        'value' => t('..'),
        'folder_path_data_name' => $folder_path_data_name,
        'folder_path' => $parent_folder_path,
        'file_browser_html_id' => $file_browser_html_id,
        'field' => $variables['field'],
        'language' => $variables['language'],
      );

      $table_rows[] = array(
        'data' => array(
          'entry_name' => array(
            'data' => _gitlab_field_widget_file_browser_folder_button($button_params),
            'colspan' => array(
              '#markup' => 2,
            ),
          ),
        ),
      );
    }

    $linked_files = array();
    // Collect linked files according to the current folder path.
    foreach ($variables['linked_files'] as $file) {
      $cond_1 = ($file->account_id == $variables['account']->remote_id);
      $cond_2 = ($file->project_id == $variables['project']['id']);
      $cond_3 = ($file->branch_id == $variables['branch']['name']);
      $cond_4 = ($file->folder_path == $variables['folder_path']);
      if ($cond_1 && $cond_2 && $cond_3 && $cond_4) {
        $linked_files[] = $file->name;
      }
    }

    foreach ($variables['files_in_folder'] as $file) {
      if ($file['type'] == 'tree') {
        $button_params = array(
          'value' => $file['name'],
          'folder_path_data_name' => $folder_path_data_name,
          'folder_path' => empty($variables['folder_path']) ? $file['name'] : $variables['folder_path'] . '/' . $file['name'],
          'file_browser_html_id' => $file_browser_html_id,
          'field' => $variables['field'],
          'language' => $variables['language'],
        );

        $table_rows[] = array(
          'data' => array(
            'entry_name' => _gitlab_field_widget_file_browser_folder_button($button_params),
            'file_name' => array(
              '#markup' => '&nbsp;',
            ),
          ),
        );
      }
      else {
        if (in_array($file['name'], $linked_files)) {
          $table_rows[] = array(
            'data' => array(
              'entry_name' => array(
                '#markup' => check_plain($file['name']),
              ),
              'file_name' => array(
                '#type' => 'checkbox',
                '#attributes' => array(
                  'checked' => 'checked',
                ),
                '#disabled' => TRUE,
              ),
            ),
          );
        }
        else {
          $linkable_files_available = TRUE;
          $table_rows[] = array(
            'data' => array(
              'entry_name' => array(
                '#markup' => check_plain($file['name']),
              ),
              'file_name' => array(
                '#type' => 'checkbox',
                '#return_value' => $file['name'],
              ),
            ),
          );
        }
      }
    }
  }
  else {
    if (!empty($variables['account']) && !empty($variables['project']) && !empty($variables['branch'])) {
      $table_empty_message = t("The requested path doesn't exist or the project repository is empty.");
    }
  }

  // File browser container.
  $file_browser = array(
    '#type' => 'container',
    '#attributes' => array(
      'id' => $file_browser_html_id,
      'class' => array('file-browser'),
    ),
    'messages' => array(
      '#type' => 'container',
      '#attributes' => array(
        'id' => $variables['field']['field_name'] . '_gitlab_field_widget_file_browser_messages',
      ),
    ),
    'breadcrumb' => array(
      '#theme' => 'gitlab_field_widget_file_browser_breadcrumb',
      'buttons' => $breadcrumb_button_elements,
    ),
    'table' => array(
      '#theme' => 'gitlab_field_widget_file_table',
      '#header' => array(
        t('File name'),
        array('data' => t('Link'), 'width' => '20'),
      ),
      'rows' => $table_rows,
      '#caption' => NULL,
      '#colgroups' => array(),
      '#sticky' => FALSE,
      '#empty' => $table_empty_message,
    ),
  );

  // Add a "Link files" button if there are one or more linkable files.
  if ($linkable_files_available) {
    $file_browser['link_files'] = array(
      '#type' => 'button',
      '#value' => t('Link selected files'),
      '#name' => $variables['field']['field_name'] . '_gitlab_field_widget_link_files_button',
      '#link_new_files' => TRUE,
      // Only validate the field the widget belongs to.
      '#limit_validation_errors' => array(array($variables['field']['field_name'])),
      '#attributes' => array(
        $folder_path_data_name => $folder_path,
      ),
      '#widget_refresh' => array(
        'field' => array(
          'name' => $variables['field']['field_name'],
          'language' => $variables['language'],
        ),
        'replacements' => array(
          array(
            'name_path' => 'linked_files',
            'html_id' => $variables['field']['field_name'] . '_gitlab_field_widget_linked_files'
          ),
          array(
            'name_path' => 'link_new_files][file_browser',
            'html_id' => $file_browser_html_id
          ),
        ),
      ),
      '#ajax' => array(
        'callback' => '_gitlab_field_widget_refresh',
        'effect' => 'fade',
      ),
    );
  }

  return $file_browser;
}

/**
 * Constructs a render array for an AJAX folder button element.
 *
 * @param array $variables
 *   An associative array containing:
 *   - value: The value of the button.
 *   - folder_path_data_name: The HTML name of folder path's data-* attribute.
 *   - folder_path: The folder's path the button pints to.
 *   - file_browser_html_id: The HTML id of the file browser container.
 *   - field: An array of the field the file browser widget belongs to.
 *   - language: The language associated with widget items.
 *
 * @return array
 *   Returns a render array of the button element.
 *
 * @see gitlab_field_widget_get_file_browser()
 */
function _gitlab_field_widget_file_browser_folder_button(array $variables) {
  // Create a static index variable for unique button names.
  static $button_index = 0;

  return array(
    '#type' => 'button',
    '#value' => check_plain($variables['value']),
    '#name' => $variables['field']['field_name'] . '_gitlab_field_widget_folder_button_' . $button_index++,
    // Only validate the field the widget belongs to.
    '#limit_validation_errors' => array(array($variables['field']['field_name'])),
    '#attributes' => array(
      $variables['folder_path_data_name'] => $variables['folder_path'],
      'class' => array('file-browser-folder'),
    ),
    '#link_new_files' => TRUE,
    '#widget_refresh' => array(
      'field' => array(
        'name' => $variables['field']['field_name'],
        'language' => $variables['language'],
      ),
      'replacements' => array(
        array(
          'name_path' => 'link_new_files][file_browser',
          'html_id' => $variables['file_browser_html_id']
        ),
      ),
    ),
    '#ajax' => array(
      'callback' => '_gitlab_field_widget_refresh',
      'effect' => 'fade',
    ),
  );
}

/**
 * AJAX callback to refresh the widget.
 *
 * @see gitlab_field_field_widget_form()
 */
function _gitlab_field_widget_refresh($form, $form_state) {
  $commands = array();
  if (!empty($form_state['triggering_element']['#widget_refresh'])) {
    $info = $form_state['triggering_element']['#widget_refresh'];

    $form_errors = array(
      'linked_files' => array(
        'messages' => _gitlab_field_widget_get_message($info['field']['name'], 'error', 'linked_files'),
        'messages_html_id' => $info['field']['name'] . '_gitlab_field_widget_linked_files_messages',
      ),
      'file_sources' => array(
        'messages' => _gitlab_field_widget_get_message($info['field']['name'], 'error', 'file_sources'),
        'messages_html_id' => $info['field']['name'] . '_gitlab_field_widget_file_sources_messages',
      ),
      'file_browser' => array(
        'messages' => _gitlab_field_widget_get_message($info['field']['name'], 'error', 'file_browser'),
        'messages_html_id' => $info['field']['name'] . '_gitlab_field_widget_file_browser_messages',
      ),
    );

    // Check form errors.
    $form_has_errors = FALSE;
    foreach ($form_errors as $errors) {
      if (!empty($errors['messages'])) {
        $form_has_errors = TRUE;
        $themed_messages = theme('gitlab_field_widget_messages', array(
          'messages' => $errors['messages'],
          'type' => 'error',
        ));
        $commands[] = ajax_command_html('#' . $errors['messages_html_id'], $themed_messages);
      }
    }

    if (!$form_has_errors) {
      foreach ($info['replacements'] as $replacement) {
        $element = $form[$info['field']['name']][$info['field']['language']];
        $path_elements = explode('][', $replacement['name_path']);
        foreach ($path_elements as $key) {
          if (isset($element[$key])) {
            $element = $element[$key];
          }
        }
        $commands[] = ajax_command_replace('#' . $replacement['html_id'], drupal_render($element));
      }

      // Check for changes and show the warning message if any.
      $changes_warning = _gitlab_field_widget_has_changes_warning($info['field']['name'], $form_state);

      if ($changes_warning) {
        $changes_warning_html_id = $info['field']['name'] . '_gitlab_field_widget_changes_warning';
        $commands[] = ajax_command_invoke('#' . $changes_warning_html_id, 'addClass', array('visible'));
        $commands[] = ajax_command_invoke('#' . $changes_warning_html_id, 'fadeIn', array('slow'));
      }
    }
  }
  return array(
    '#type' => 'ajax',
    '#commands' => $commands,
  );
}

/**
 * Helper function to set a value in $form_state for later usage.
 *
 * @param $field_name
 *   The field name the widget belongs to.
 * @param $form_state
 *   An associative array containing the current state of the form.
 * @param $key
 *   The key under the value should be stored.
 * @param $value
 *   The value which should be stored.
 */
function _gitlab_field_widget_set_value($field_name, &$form_state, $key, $value) {
  $form_state_key = $field_name . '_gitlab_field_widget_values';
  if (!isset($form_state[$form_state_key])) {
    $form_state[$form_state_key] = array();
  }
  $form_state[$form_state_key][$key] = $value;
}

/**
 * Helper function to get a previously saved value from $form_state.
 *
 * @param $field_name
 *   The field name the widget belongs to.
 * @param $form_state
 *   An associative array containing the current state of the form.
 * @param $key
 *   The key under the value should be stored.
 *
 * @return mixed
 *   Returns the value according to the specified key or NULL if no value found.
 */
function _gitlab_field_widget_get_value($field_name, $form_state, $key) {
  $form_state_key = $field_name . '_gitlab_field_widget_values';
  return empty($form_state[$form_state_key][$key]) ? NULL : $form_state[$form_state_key][$key];
}

/**
 * Helper function to set a file field widget related message.
 *
 * @param $field_name
 *   The field name the widget belongs to.
 * @param $type
 *   The type of the message.
 * @param $key
 *   The key under the message should be stored.
 * @param $message
 *   The message which should be stored.
 */
function _gitlab_field_widget_set_message($field_name, $type, $key, $message) {
  $messages = &drupal_static($field_name . '_gitlab_field_widget_messages');
  if (!isset($messages)) {
    $messages = array();
  }
  if (!isset($messages[$type])) {
    $messages[$type] = array();
  }
  if (!isset($messages[$type][$key])) {
    $messages[$type][$key] = array();
  }
  $messages[$type][$key][] = $message;
}

/**
 * Helper function to get a file field widget related message.
 *
 * @param $field_name
 *   The field name the widget belongs to.
 * @param $type
 *   The type of the message.
 * @param $key
 *   The key under the message should be stored.
 *
 * @returns mixed
 *   Returns with the specified message or NULL if it not found.
 */
function _gitlab_field_widget_get_message($field_name, $type, $key) {
  $messages = &drupal_static($field_name . '_gitlab_field_widget_messages');
  return empty($messages[$type][$key]) ? NULL : $messages[$type][$key];
}

/**
 * Set a boolean to show a warning message about temporarily changes.
 *
 * @param $field_name
 *   The machine name of the field the widget belongs to.
 * @param $form_state
 *   An associative array containing the current state of the form.
 */
function _gitlab_field_widget_set_changes_warning($field_name, &$form_state) {
  $form_state[$field_name . '_gitlab_field_widget_changes_warning'] = TRUE;
}

/**
 * Get a boolean to show a warning message about temporarily changes.
 *
 * @param $field_name
 *   The machine name of the field the widget belongs to.
 * @param $form_state
 *   An associative array containing the current state of the form.
 *
 * @returns boolean
 *   Returns TRUE if the warning message needs to be shown. Returns FALSE
 *   otherwise.
 */
function _gitlab_field_widget_has_changes_warning($field_name, &$form_state) {
  return !empty($form_state[$field_name . '_gitlab_field_widget_changes_warning']);
}

/**
 * Helper function which returns account's projects as id => name assoc. array.
 *
 * @param \GitLabAccountEntity $account
 *   Gitlab account object.
 *
 * @return array
 *   Array of projects.
 */
function _gitlab_field_get_project_field_options(GitLabAccountEntity $account) {
  $project_options = array();
  $projects = gitlab_projects_index($account->getPrivateToken(), array(
    'wsdata' => array(
      'endpoint' => $account->endpoint_url,
    ),
  ));
  if (!empty($projects['data'])) {
    foreach ($projects['data'] as $_project) {
      $project_options[$_project['id']] = $_project['name_with_namespace'];
    }
  }
  return $project_options;
}

/**
 * Helper function which returns branches in the account's given project.
 *
 * @param \GitLabAccountEntity $account
 *   Gitlab account object.
 * @param int $project_id
 *   Project's id on Gitlab.
 *
 * @return array
 *   Array of branches.
 */
function _gitlab_field_get_branch_field_options(GitLabAccountEntity $account, $project_id) {
  $branch_options = array();
  $branches = gitlab_projects_branches_index($project_id, $account->getPrivateToken(), array(
    'wsdata' => array(
      'endpoint' => $account->endpoint_url,
    ),
  ));
  if (!empty($branches['data'])) {
    foreach ($branches['data'] as $_branch) {
      $branch_options[$_branch['name']] = $_branch['name'];
    }
  }

  return $branch_options;
}
